#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Session wrap by sacrud_sessionmaker tests
"""

import unittest

import transaction
from sacrud import CRUDSession, crud_sessionmaker
from zope.sqlalchemy import ZopeTransactionExtension

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker, scoped_session

from . import User, Profile


class SessionMethodTest(unittest.TestCase):

    def test_session(self):
        Session = scoped_session(sessionmaker(class_=CRUDSession))
        session = Session()
        self.assertEqual(hasattr(session, 'sacrud'), True)
        self.assertEqual(hasattr(session, 'sacrud'), True)


class _ActionTest(object):

    zope = None

    def _add_item(self, table, *args, **kwargs):
        obj = table(*args, **kwargs)
        self.session.add(obj)
        if self.zope:
            transaction.commit()
        else:
            self.session.commit()
        return obj

    def test_create(self):
        user = self.session.sacrud(User)\
            .create({'name': 'Dzhirad', 'fullname': 'Kiri', 'password': 123})
        if not self.zope:
            self.session.commit()
        self.assertEqual(user.name, 'Dzhirad')
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(user.id, db_user.id)

    def test_read(self):
        self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.sacrud(User).read(1)
        self.assertEqual(user.name, 'Vasya')
        self.assertEqual(user.id, 1)

    def test_update(self):
        self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.sacrud(User).update(1, {'name': 'Bill'})
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(user.name, 'Bill')
        self.assertEqual(user.id, db_user.id)

    def test_delete(self):
        user = self._add_item(User, 'Volod', 'Khonin', "123")
        self.session.sacrud(User).delete(user.id)
        db_user = self.session.query(User).filter_by(id=user.id).all()
        self.assertEqual(db_user, [])


class ZopeSessionTest(_ActionTest, unittest.TestCase):

    zope = True

    def setUp(self):
        engine = create_engine('sqlite:///:memory:')
        self.session = crud_sessionmaker(scoped_session(
            sessionmaker(extension=ZopeTransactionExtension(),
                         expire_on_commit=False)))
        self.session.remove()
        self.session.configure(bind=engine)

        # Create tables
        User.metadata.create_all(engine)
        Profile.metadata.create_all(engine)


class SQLAlchemySessionTest(_ActionTest, unittest.TestCase):

    zope = False

    def setUp(self):
        engine = create_engine('sqlite:///:memory:')
        Session = scoped_session(
            sessionmaker(class_=CRUDSession, expire_on_commit=False))
        Session.remove()
        Session.configure(bind=engine)
        self.session = Session()

        # Create tables
        User.metadata.create_all(engine)
        Profile.metadata.create_all(engine)



# -*- coding: utf-8 -*-
"""
action.py tests
"""

import os
import unittest

import transaction
from sacrud.common import TableProperty
from zope.sqlalchemy import ZopeTransactionExtension

from sqlalchemy import orm, Table, create_engine
from sqlalchemy.orm import mapper, backref, relationship
from sqlalchemy.types import LargeBinary as BYTEA
from sqlalchemy.types import (
    Date,
    Enum,
    Text,
    Float,
    String,
    Boolean,
    Integer,
    DateTime,
    TIMESTAMP
)
from sqlalchemy.schema import Column, ForeignKey
from sqlalchemy.ext.declarative import declarative_base

Base = declarative_base()

DIRNAME = os.path.dirname(__file__)
DB_FILE = os.path.join(DIRNAME, 'test.sqlite')
TEST_DATABASE_CONNECTION_STRING = 'sqlite:///%s' % DB_FILE


class MockCGIFieldStorage(object):
    pass


class BaseTest(unittest.TestCase):

    def setUp(self):
        self.engine = create_engine('sqlite:///:memory:')
        self.session2 = orm.scoped_session(orm.sessionmaker())
        self.session2.remove()
        self.session2.configure(bind=self.engine)

    def tearDown(self):
        pass


class BaseZopeTest(BaseTest):

    zope = True

    def _add_item(self, table, *args, **kwargs):
        obj = table(*args, **kwargs)
        self.session.add(obj)
        transaction.commit()
        return obj

    def setUp(self):
        super(BaseZopeTest, self).setUp()
        self.session = orm.scoped_session(
            orm.sessionmaker(extension=ZopeTransactionExtension(),
                             expire_on_commit=False))
        self.session.remove()
        self.session.configure(bind=self.engine)

        # Create tables
        User.metadata.create_all(self.engine)
        Profile.metadata.create_all(self.engine)


class BaseSQLAlchemyTest(BaseTest):

    zope = False

    def _add_item(self, table, *args, **kwargs):
        obj = table(*args, **kwargs)
        self.session.add(obj)
        self.session.commit()
        return obj

    def setUp(self):
        super(BaseSQLAlchemyTest, self).setUp()
        self.session = orm.scoped_session(orm.sessionmaker())
        self.session.remove()
        self.session.configure(bind=self.engine)

        # Create tables
        User.metadata.create_all(self.engine)
        Profile.metadata.create_all(self.engine)


association_table = Table('association', Base.metadata,
                          Column('group_id', Integer, ForeignKey('group.id')),
                          Column('user_id', Integer, ForeignKey('user.id'))
                          )


class Groups2User(object):

    def __init__(self, group_id, user_id):
        self.group_id = group_id
        self.user_id = user_id

mapper(Groups2User, association_table,
       primary_key=[association_table.c.group_id, association_table.c.user_id])


class Groups(Base):
    __tablename__ = 'group'
    id = Column(Integer, primary_key=True)
    name = Column(String)
    users = relationship("User", secondary=association_table)


class MultiPK(Base):

    __tablename__ = 'multipk'

    id = Column(Integer, primary_key=True)
    id2 = Column(Integer, primary_key=True)
    id3 = Column(Integer, primary_key=True)

    fk = Column('group_id', Integer, ForeignKey('group.id'))


class TypesPreprocessor(Base):
    __tablename__ = 'types_preprocessor'
    id = Column(Integer, primary_key=True)
    sak = Column(BYTEA, nullable=False)
    date = Column(Date)
    datetime = Column(DateTime)
    datetimeseconds = Column(TIMESTAMP)


class User(Base):

    __tablename__ = 'user'

    id = Column(Integer, primary_key=True)
    name = Column(String, default='')
    fullname = Column(String)
    password = Column(String, info={'verbose_name': 'user password'})
    age = Column(Integer)
    sex = Column(Enum('male',
                      'female',
                      'alien',
                      'unknown', name="sex"))
    type = Column(String(50))

    groups = relationship("Groups", secondary=association_table)

    __mapper_args__ = {
        'polymorphic_on': type,
        'polymorphic_identity': 'users',
    }

    sacrud_detail_col = [('', (name, fullname, password)),
                         ('other', (sex,))]

    @TableProperty
    def foo(self):
        return "I'm property"

    def __init__(self, name, fullname, password, sex='unknown', groups=[]):
        self.name = name
        self.age = 0
        self.fullname = fullname
        self.password = password
        self.sex = sex
        self.groups = groups


class Profile(Base):

    __tablename__ = 'profile'

    id = Column(Integer, primary_key=True)
    user_id = Column(Integer, ForeignKey("user.id"), nullable=False)
    user = relationship(User, backref=backref("profile", lazy="joined"))
    phone = Column(String)
    cv = Column(Text)
    married = Column(Boolean)
    salary = Column(Float)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.
import json

import transaction
from sacrud.action import CRUD

import sqlalchemy

from . import (
    User,
    Groups,
    Profile,
    Groups2User,
    BaseZopeTest,
    BaseSQLAlchemyTest
)


class ReadTest(object):

    def test_read(self):
        user = CRUD(self.session, User).read().all()
        self.assertEqual(len(user), 20)

    def test_read_single(self):
        user = CRUD(self.session, User).read(5)
        self.assertEqual(user.id, 5)

    def test_read_by_str_id(self):
        user = CRUD(self.session, User).read('5')
        self.assertEqual(user.id, 5)

    def test_read_args(self):
        users = CRUD(self.session, User).read(1, '2', 3, 10, 20)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_of_tuple(self):
        tuple_of_users = (1, 2, 3, 10, 20)
        users = CRUD(self.session, User).read(tuple_of_users)
        self.assertEqual([u.id for u in users], list(tuple_of_users))

    def test_read_of_list(self):
        list_of_users = [1, 2, 3, 10, 20]
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], list_of_users)

    def test_read_json_of_list(self):
        list_of_users = "[1, 2, 3, 10, 20]"
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_tuple_of_str(self):
        tuple_of_users = ('1', '2', 3, '10', '20')
        users = CRUD(self.session, User).read(tuple_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_list_of_str(self):
        list_of_users = ['1', '2', '3', '10', '20']
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_json_list_of_str(self):
        list_of_users = '["1", "2", "3", "10", "20"]'
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_by_dict_pk(self):
        user = CRUD(self.session, User).read({'id': 19})
        self.assertEqual(user.one().id, 19)

    def test_read_json_by_dict_pk(self):
        user = CRUD(self.session, User).read('{"id": 19}')
        self.assertEqual(user.one().id, 19)

    def test_read_list_of_dict(self):
        list_of_users = [{'id': '1'}, {'id': '2'}, {'id': '3'},
                         {'id': 10}, {'id': 20}]
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_json_list_of_dict(self):
        list_of_users = '''[{"id": "1"}, {"id": "2"}, {"id": "3"},
                            {"id": 10}, {"id": 20}]'''
        users = CRUD(self.session, User).read(list_of_users)
        self.assertEqual([u.id for u in users], [1, 2, 3, 10, 20])

    def test_read_list_of_dict_and_then_delete(self):
        list_of_users = [{'id': '1'}, {'id': '2'}, {'id': '3'},
                         {'id': 10}, {'id': 20}]
        CRUD(self.session, User).read(list_of_users)\
            .delete(synchronize_session=False)
        rows = [item.id for item in CRUD(self.session, User).read().all()]
        self.assertNotIn(1, rows)
        self.assertNotIn(2, rows)
        self.assertNotIn(3, rows)
        self.assertNotIn(10, rows)
        self.assertNotIn(20, rows)


class CreateTest(object):

    def test_create(self):
        self.request['name'] = 'foo'
        group = CRUD(self.session, Groups).create(self.request)
        self.assertEqual(group.name, 'foo')

        db_group = self.session.query(Groups).get(group.id)
        self.assertEqual(group.id, db_group.id)

    def test_json_create(self):
        self.request['name'] = 'foo'
        group = CRUD(self.session, Groups).create(json.dumps(self.request))
        self.assertEqual(group.name, 'foo')

        db_group = self.session.query(Groups).get(group.id)
        self.assertEqual(group.id, db_group.id)

    def test_create_twice(self):
        self.request['name'] = 'foo'
        group = CRUD(self.session, Groups).create(self.request)
        self.assertEqual(group.name, 'foo')

        if self.zope:
            transaction.commit()
        else:
            self.session.commit()

        self.request['id'] = group.id
        try:
            CRUD(self.session, Groups).create(self.request)
        except Exception as e:
            if self.zope:
                from sqlalchemy.exc import IntegrityError
                transaction.abort()
                self.assertEqual(IntegrityError, type(e))
            else:
                from sqlalchemy.orm.exc import FlushError
                self.session.rollback()
                self.assertEqual(FlushError, type(e))

    def test_create_twice_with_update_options(self):
        self.request['name'] = 'foo'
        group = CRUD(self.session, Groups).create(self.request)
        self.assertEqual(group.name, 'foo')

        if self.zope:
            transaction.commit()
        else:
            self.session.commit()

        self.request['id'] = group.id
        self.request['name'] = 'new_foo'
        group = CRUD(self.session, Groups).create(self.request, update=True)

        if self.zope:
            transaction.commit()
        else:
            self.session.commit()

        self.assertEqual(group.name, 'new_foo')

    def test_create_with_empty_post_request(self):
        self.request = {}
        group = CRUD(self.session, Groups).create(self.request)
        self.assertEqual(group.id, 1)

        db_group = self.session.query(Groups).get(group.id)
        self.assertEqual(group.id, db_group.id)

    def test_create_json_with_empty_post_request(self):
        group = CRUD(self.session, Groups).create("{}")
        self.assertEqual(group.id, 1)

        db_group = self.session.query(Groups).get(group.id)
        self.assertEqual(group.id, db_group.id)

    def test_create_composit_pk(self):
        CRUD(self.session, Groups2User).create({'group_id': 2, 'user_id': 1})
        obj = self.session.query(Groups2User).filter_by(group_id=2)\
            .filter_by(user_id=1).one()
        self.assertEqual(obj.group_id, 2)
        self.assertEqual(obj.user_id, 1)

    def test_create_json_composit_pk(self):
        CRUD(self.session, Groups2User).create('{"group_id": 2, "user_id": 1}')
        obj = self.session.query(Groups2User).filter_by(group_id=2)\
            .filter_by(user_id=1).one()
        self.assertEqual(obj.group_id, 2)
        self.assertEqual(obj.user_id, 1)

    def test_create_user(self):
        # Create users
        request = {}
        request['name'] = ["Vasya", ]
        request['fullname'] = ["Vasya Pupkin", ]
        request['password'] = ["", ]  # check empty value
        request['groups[]'] = [u'["id", 1]', u'["id", 3]', u'["id" bad row]']
        request['badAttr'] = ["1", "bar"]
        request['badM2MAttr[]'] = ["1", "bar"]
        request['groups[]'] = None
        CRUD(self.session, User).create(request)
        user = self.session.query(User).get(1)
        self.assertEqual([x.id for x in user.groups], [])

        # Add profile
        request = {}
        request['phone'] = ["213123123", ]
        request['cv'] = ["Vasya Pupkin was born in Moscow", ]
        request['married'] = ["true", ]
        request["salary"] = ["23.0", ]
        request["user_id"] = ["1", ]

        CRUD(self.session, Profile).create(request)

        profile = self.session.query(Profile).get(1)

        self.assertEqual(profile.phone, "213123123")
        self.assertEqual(profile.cv, "Vasya Pupkin was born in Moscow")
        self.assertEqual(profile.married, True)
        self.assertEqual(float(profile.salary), float(23))
        self.assertEqual(profile.user.id, 1)

    def test_create_with_relation(self):
        # Create groups (M2M example)
        request = {}
        request['name'] = 'foo'
        group1 = CRUD(self.session, Groups).create(request)
        CRUD(self.session, Groups).create(request)
        group3 = CRUD(self.session, Groups).create(request)

        # Create users
        request = {}
        request['name'] = ["Vasya", ]
        request['fullname'] = ["Vasya Pupkin", ]
        request['password'] = ["", ]  # check empty value
        request['groups[]'] = [u'["id", 1]', u'["id", 3]', u'["id" bad row]']
        request['badAttr'] = ["1", "bar"]
        request['badM2MAttr[]'] = ["1", "bar"]

        user = CRUD(self.session, User).create(request)
        user = self.session.query(User).get(user.id)

        self.assertEqual(user.name, "Vasya")
        self.assertEqual(user.fullname, "Vasya Pupkin")
        self.assertEqual(user.password, '')
        self.assertEqual([x.id for x in user.groups],
                         [group1.id, group3.id])

    def test_create_no_commit(self):
        self.request['name'] = 'foo'
        group = CRUD(self.session, Groups).create(self.request, commit=False)
        self.assertEqual(group.name, 'foo')
        self.assertEqual(group.id, None)
        if self.zope:
            transaction.commit()
        else:
            self.session.commit()
        db_group = self.session.query(Groups).get(group.id)
        self.assertEqual(group.id, db_group.id)

    def test_create_m2m(self):
        self._add_item(Groups, name='foo')
        self._add_item(Groups, name='bar')
        self._add_item(Groups, name='baz')

        self.request['name'] = 'foo'
        self.request['fullname'] = 'foo'
        self.request['password'] = 'foo'
        self.request['groups[]'] = ['["id", 1]', '["id", 2]']
        app = CRUD(self.session, User).create(self.request)
        self.assertEqual([g.id for g in app.groups], [1, 2])


class UpdateTest(object):

    def test_update_by_int_id(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session, User).update(user.id, data={'name': 'Petya'})
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(db_user.name, 'Petya')

    def test_update_from_0_to_null(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', '123')
        user = self.session.query(User).get(user.id)
        self.assertEqual(user.age, 0)

        CRUD(self.session, User).update(user.id,
                                        data={'age': sqlalchemy.sql.null()})
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(db_user.age, None)

    def test_update_json_by_int_id(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session, User).update(user.id, data='{"name": "Petya"}')
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(db_user.name, 'Petya')

    def test_update_by_str_id(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session, User).update(str(user.id), data={'name': 'Petya'})
        self.assertEqual(user.name, 'Petya')

    def test_update_by_composit_pk(self):
        self._add_item(Groups2User, **{'group_id': 1, 'user_id': 1})
        self._add_item(Groups2User, **{'group_id': 2, 'user_id': 1})
        self._add_item(Groups2User, **{'group_id': 3, 'user_id': 2})
        self._add_item(Groups2User, **{'group_id': 5, 'user_id': 1})

        CRUD(self.session, Groups2User)\
            .update({'group_id': 2, 'user_id': 1}, data={'group_id': '10'})
        obj = self.session.query(Groups2User).filter_by(group_id=10)\
            .filter_by(user_id=1).one()
        self.assertEqual(obj.group_id, 10)
        self.assertEqual(obj.user_id, 1)

    def test_update_json_by_composit_pk(self):
        self._add_item(Groups2User, **{'group_id': 1, 'user_id': 1})
        self._add_item(Groups2User, **{'group_id': 2, 'user_id': 1})
        self._add_item(Groups2User, **{'group_id': 3, 'user_id': 2})
        self._add_item(Groups2User, **{'group_id': 5, 'user_id': 1})

        CRUD(self.session, Groups2User)\
            .update('{"group_id": 2, "user_id": 1}', data='{"group_id": "10"}')
        obj = self.session.query(Groups2User).filter_by(group_id=10)\
            .filter_by(user_id=1).one()
        self.assertEqual(obj.group_id, 10)
        self.assertEqual(obj.user_id, 1)

    def test_update_profile(self):

        # Add user 1
        user1 = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user1 = self.session.query(User).get(user1.id)
        profile = self._add_item(Profile, user=user1, salary="25.7")

        # Add user 2
        user2 = self._add_item(User, 'Vasya', 'Pupkin', "123")
        self.session.query(User).get(user2.id)

        profile = self.session.query(Profile).get(profile.id)
        request = {}
        request['phone'] = ["213123123", ]
        request['cv'] = ["Vasya Pupkin was born in Moscow", ]
        request['married'] = ["true", ]
        request["salary"] = ["23.0", ]
        request["user_id"] = ["2", ]

        CRUD(self.session, Profile).update(pk={'id': profile.id}, data=request)
        profile = self.session.query(Profile).get(profile.id)

        self.assertEqual(profile.phone, "213123123")
        self.assertEqual(profile.cv, "Vasya Pupkin was born in Moscow")
        self.assertEqual(profile.married, True)
        self.assertEqual(profile.user.id, 2)
        self.assertEqual(float(profile.salary), float(23))

    def test_update_no_commit(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session2, User).update(user.id, data={'name': 'Petya'})
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(db_user.name, 'Vasya')

        self.session2.commit()
        self.session.close()
        db_user = self.session.query(User).get(user.id)
        self.assertEqual(db_user.name, 'Petya')


class DeleteTest(object):

    def test_delete_by_int_id(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session, User).delete(user.id)
        user = self.session.query(User).filter_by(id=user.id).all()
        self.assertEqual(user, [])

    def test_delete_by_str_id(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session, User).delete(str(user.id))
        user = self.session.query(User).filter_by(id=user.id).all()
        self.assertEqual(user, [])

    def test_delete_by_composit_pk(self):
        self._add_item(Groups2User, **{'group_id': 3, 'user_id': 2})

        CRUD(self.session, Groups2User).delete({'group_id': 3, 'user_id': 2})
        obj = self.session.query(Groups2User).filter_by(group_id=3)\
            .filter_by(user_id=2).all()
        self.assertEqual(obj, [])

    def test_delete_json_by_composit_pk(self):
        self._add_item(Groups2User, **{'group_id': 3, 'user_id': 2})

        CRUD(self.session, Groups2User).delete('{"group_id": 3, "user_id": 2}')
        obj = self.session.query(Groups2User).filter_by(group_id=3)\
            .filter_by(user_id=2).all()
        self.assertEqual(obj, [])

    def test_delete_user(self):

        user = User(u'Vasya', u'Pupkin', u"123")
        self.session.add(user)
        if self.zope:
            transaction.commit()
        else:
            self.session.commit()

        request = {}
        request['phone'] = ["213123123", ]
        request['cv'] = ["Vasya Pupkin was born in Moscow", ]
        request['married'] = ["true", ]
        request["salary"] = ["23.0", ]
        request["user_id"] = ["1", ]

        user = CRUD(self.session, Profile).create(request)
        profile = CRUD(self.session, Profile).delete({'id': user.id})

        profile = self.session.query(Profile).get(profile['pk']['id'])
        self.assertEqual(profile, None)

    def test_delete_no_commit(self):
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        user = self.session.query(User).get(user.id)

        CRUD(self.session2, User).delete(user.id, commit=False)
        users = self.session.query(User).filter_by(id=user.id).all()
        self.assertEqual(len(users), 1)

        self.session2.commit()
        user = self.session.query(User).filter_by(id=user.id).all()
        self.assertEqual(len(user), 0)


class ZopeTransaction(BaseZopeTest, ReadTest, CreateTest, UpdateTest,
                      DeleteTest):

    def setUp(self):
        super(ZopeTransaction, self).setUp()
        self.request = {}

        # Generator of Users
        users = [User(i, i, i) for i in range(20)]
        self.session.add_all(users)


class SQLAlchemyTransaction(BaseSQLAlchemyTest, ReadTest, CreateTest,
                            UpdateTest, DeleteTest):

    def setUp(self):
        super(SQLAlchemyTransaction, self).setUp()
        self.request = {}

        # Generator of Users
        users = [User(i, i, i) for i in range(20)]
        self.session.add_all(users)



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Test for sacrud.common.sa_helpers
"""
import unittest

from sacrud.common import (
    get_pk,
    unjson,
    get_obj,
    pk_to_list,
    pk_list_to_dict,
    columns_by_group,
    get_flat_columns,
    get_relationship,
    get_empty_instance
)

from . import (
    User,
    Groups,
    MultiPK,
    Profile,
    BaseZopeTest,
    association_table,
    BaseSQLAlchemyTest
)


class CommonTest(unittest.TestCase):

    def test_unjson(self):
        foo = unjson(1)
        self.assertEqual(foo, 1)

    def test_unjson_json(self):
        foo = unjson('{"foo": 22}')
        self.assertEqual(foo, {'foo': 22})


class SQLAlchemyHelpersTest(object):

    def test_get_obj_with_bad_pk(self):
        self.assertEqual(None, get_obj(self.session, User, None))
        self.assertEqual(None, get_obj(self.session, User, ''))
        self.assertEqual(None, get_obj(self.session, User, {}))

    def test_get_pk(self):
        # class
        pk = get_pk(User)
        self.assertEqual('id', pk[0].name)

        # object
        user = self._add_item(User, 'Vasya', 'Pupkin', "123")
        pk = get_pk(user)
        self.assertEqual('id', pk[0].name)

    def test_table_property(self):
        user = User('Vasya', 'Pupkin', '123')
        self.assertEqual(user.foo, "I'm property")

    def test_pk_to_list(self):
        user = User('Vasya', 'Pupkin', '123')
        self.assertEqual(pk_to_list(user), ['id', None])
        self.assertEqual(pk_to_list(user, as_json=True),
                         '["id", null]')
        user.id = 100500
        self.assertEqual(pk_to_list(user), ['id', 100500])

        multipk = MultiPK(id=1, id2=1, id3=2, fk=1)
        self.assertEqual(pk_to_list(multipk),
                         ['id', 1, 'id2', 1, 'id3', 2])

        # JSON
        self.assertEqual(pk_to_list(multipk, as_json=True),
                         '["id", 1, "id2", 1, "id3", 2]')

    def test_get_relationship(self):
        foo = get_relationship(User)
        self.assertEqual(len(foo), 2)
        self.assertIn(association_table.c.user_id, foo[0].remote_side)
        self.assertIn(association_table.c.group_id, foo[0].remote_side)
        self.assertIn(Profile.__table__.c.user_id, foo[1].remote_side)

        bar = get_relationship(None)
        self.assertEqual(bar, None)

        baz = get_relationship(Groups)
        self.assertEqual(len(baz), 1)
        self.assertIn(association_table.c.user_id, foo[0].remote_side)
        self.assertIn(association_table.c.group_id, foo[0].remote_side)

    def test_pk_list_to_dict(self):
        l = ['id', 1, 'id2', 22, 'foo', 'bar']
        d = pk_list_to_dict(l)
        self.assertEqual(d, {'foo': 'bar', 'id2': 22, 'id': 1})

    def test_pk_list_to_dict2(self):
        foo = [1, 2, 3, "a", "b", {"foo": "bar"}]
        resp = pk_list_to_dict(foo)
        self.assertEqual(resp, {1: 2, 3: 'a', 'b': {'foo': 'bar'}})

        foo = [1, 2, 3, "a", "b"]
        resp = pk_list_to_dict(foo)
        self.assertEqual(resp, None)

    def test_bad_pk_list_to_dict(self):
        l = ['id', 1, 'id2', 22, 'foo']
        d = pk_list_to_dict(l)
        self.assertEqual(d, None)

    def test_columns_by_groups(self):
        c = columns_by_group(User)
        self.maxDiff = None
        table = User.__table__
        self.assertEqual(c, [('', (table.c.name, table.c.fullname,
                                   table.c.password)),
                             ('other', (table.c.sex,))]
                         )
        c = columns_by_group(Groups)
        table = Groups.__table__
        self.assertEqual(c[0][0], '')
        self.assertEqual([x for x in c[0][1]], [table.c.id, table.c.name])

    def test_get_flat_columns(self):
        c = get_flat_columns(User)
        table = User.__table__.c
        self.assertEqual(c, [table.name,
                             table.fullname,
                             table.password,
                             table.sex]
                         )

    def test_get_flat_columns_wo_settings(self):
        c = get_flat_columns(Groups)
        self.assertEqual(c, [])

    def test_get_empty_instance(self):
        ins = get_empty_instance(Profile)
        self.assertEqual(ins.id, None)


class ZopeTransaction(BaseZopeTest, SQLAlchemyHelpersTest):
    pass


class SQLAlchemyTransaction(BaseSQLAlchemyTest, SQLAlchemyHelpersTest):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Test for sacrud.common.sa_helpers
"""
import datetime

from sacrud.action import CRUD
from sacrud.preprocessing import RequestPreprocessing

from sqlalchemy import Column, String, Boolean, Integer, ForeignKey
from sqlalchemy.ext.declarative import declarative_base

from . import User, BaseZopeTest, TypesPreprocessor, BaseSQLAlchemyTest


class RequestPreprocessingTest(object):

    def test_preprocessor_hstore(self):
        prc = RequestPreprocessing({})
        foo = prc._check_hstore("{\"foo\":\"bar\"}")
        self.assertEqual(foo, {'foo': 'bar'})
        foo = prc._check_hstore("{'foo':\"bar\"}")
        self.assertEqual(foo, {'foo': 'bar'})
        with self.assertRaises(TypeError):
            prc._check_hstore("blablabla")

    def test_preprocessor_boolean_with_default_value(self):

        Base = declarative_base()

        class Foo(Base):
            __tablename__ = 'foo'

            id = Column(Integer, primary_key=True)
            visible = Column(Boolean, default=False)
            archive = Column(Boolean)

        prc = RequestPreprocessing({'visible': True})
        visible = prc.check_type(Foo, 'visible')
        self.assertEqual(visible, True)

        prc = RequestPreprocessing({'visible': False})
        visible = prc.check_type(Foo, 'visible')
        self.assertEqual(visible, False)

    def test_postgres_json_and_hstore_fields(self):

        from sqlalchemy.dialects.postgresql import JSON, JSONB, HSTORE

        Base = declarative_base()

        class Foo(Base):
            __tablename__ = 'foo'

            id = Column(Integer, primary_key=True)
            json = Column(JSON)
            jsonb = Column(JSONB)
            hstore = Column(HSTORE)

        prc = RequestPreprocessing({
            'json': '{"a": 1, "c": 3, "b": {"foo": true}}',
            'jsonb': '{"a": 1, "c": 3, "b": {"foo": true}}',
            'hstore': '{"a": 1, "c": 3, "b": 2}',
        })
        json = prc.check_type(Foo, 'json')
        jsonb = prc.check_type(Foo, 'jsonb')
        hstore = prc.check_type(Foo, 'hstore')
        self.assertEqual(json, {'a': 1, 'c': 3, 'b': {'foo': True}})
        self.assertEqual(jsonb, {'a': 1, 'c': 3, 'b': {'foo': True}})
        self.assertEqual(hstore, {'a': 1, 'c': 3, 'b': 2})

        prc = RequestPreprocessing({
            'json': '',
            'jsonb': '',
            'hstore': '',
        })
        json = prc.check_type(Foo, 'json')
        jsonb = prc.check_type(Foo, 'jsonb')
        hstore = prc.check_type(Foo, 'hstore')
        self.assertEqual(json, None)
        self.assertEqual(jsonb, None)
        self.assertEqual(hstore, None)

        with self.assertRaises(TypeError):
            prc._check_hstore("blablabla")

        prc = RequestPreprocessing({
            'json': [1, 2, 3],
            'jsonb': [1, 2, 3],
            'hstore': [1, 2, 3],
        })
        json = prc.check_type(Foo, 'json')
        jsonb = prc.check_type(Foo, 'jsonb')
        hstore = prc.check_type(Foo, 'hstore')
        self.assertEqual(json, [1, 2, 3])
        self.assertEqual(jsonb, [1, 2, 3])
        self.assertEqual(hstore, [1, 2, 3])

    def test_default_in_preprocessor(self):
        prc = RequestPreprocessing({'name': ''})
        foo = prc.check_type(User, 'name')
        self.assertEqual(foo, None)

    def test_polymorphycall_in_preprocessor(self):

        class Foo(User):
            __tablename__ = 'foo'
            __table_args__ = {'extend_existing': True}
            __mapper_args__ = {
                'polymorphic_identity': 'foo',
            }

            id = Column(Integer, ForeignKey('user.id'), primary_key=True)
            foo = Column(String)

        prc = RequestPreprocessing({'name': 'lalala'})
        name = prc.check_type(Foo, 'name')
        self.assertEqual(name, 'lalala')

    def test_preprocessor(self):
        request = {}
        request['date'] = "2012-12-12"
        request['datetime'] = "2012-12-12 12:12"
        request['datetimeseconds'] = "2012-12-12 12:12:12"
        request["sak"] = "Ac"
        foo = CRUD(self.session, TypesPreprocessor).create(request)

        self.assertEqual(foo.sak, bytearray(b"Ac"))
        # XXX: I feel dissonance here...
        if self.zope:
            self.assertEqual(foo.date, datetime.datetime(2012, 12, 12, 0, 0))
        else:
            self.assertEqual(foo.date, datetime.date(2012, 12, 12))

        self.assertEqual(foo.datetime,
                         datetime.datetime(2012, 12, 12, 12, 12))
        self.assertEqual(foo.datetimeseconds,
                         datetime.datetime(2012, 12, 12, 12, 12, 12))


class ZopeTransaction(BaseZopeTest, RequestPreprocessingTest):
    pass


class SQLAlchemyTransaction(BaseSQLAlchemyTest, RequestPreprocessingTest):
    pass



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2015 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

import types

from sqlalchemy.orm import Session

from .action import CRUD

__all__ = ['crud_sessionmaker', 'CRUDSession']


class CRUDSession(Session):
    """
    Wraps SQLAlchemy session adding there ``sacrud`` method.

    ::

        from sqlalchemy.orm import scoped_session, sessionmaker
        from sacrud import CRUDSession

        Session = scoped_session(sessionmaker(class_=CRUDSession))
        session = Session()
        DBSession.sacrud(User).delete(1)
    """
    def sacrud(self, cls, *args, **kwargs):
        return CRUD(self, cls, *args, **kwargs)


def crud_sessionmaker(session):
    """
    Wraps zope.sqlalchemy session adding there ``sacrud`` method.

    ::

        from sqlalchemy.orm import scoped_session, sessionmaker
        from zope.sqlalchemy import ZopeTransactionExtension
        from sacrud import crud_sessionmaker

        DBSession = crud_sessionmaker(scoped_session(
            sessionmaker(extension=ZopeTransactionExtension())))
        DBSession.sacrud(User).delete(1)
    """
    session.sacrud = types.MethodType(CRUD, session)
    return session



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
Preprocessing
"""
import ast
import json
import inspect
from datetime import date, datetime

import six

import sqlalchemy

from .common import get_columns


def list_of_lists_to_dict(l):
    """ Convert list of key,value lists to dict

    [['id', 1], ['id', 2], ['id', 3], ['foo': 4]]
    {'id': [1, 2, 3], 'foo': [4]}
    """
    d = {}
    for key, val in l:
        d.setdefault(key, []).append(val)
    return d


def get_m2m_objs(session, relation, id_from_request):
    mapper = relation.mapper
    pk_list = mapper.primary_key
    ids = []
    if id_from_request:
        if isinstance(id_from_request, six.string_types):
            id_from_request = [id_from_request, ]
        for id in id_from_request:
            try:
                ids.append(json.loads(id))
            except ValueError:
                pass
            except TypeError:
                return None
        ids = list_of_lists_to_dict(ids)
    else:
        ids = {}
    objs = session.query(mapper)
    for pk in pk_list:
        objs = objs.filter(pk.in_(ids.get(pk.name, [])))
    return objs.all()


def get_m2m_value(session, request, obj):
    """ Set m2m value for model obj from request params like "group[]"

        :Parameters:

            - `session`: SQLAlchemy DBSession
            - `request`: request as dict
            - `obj`: model instance
    """
    params = {}
    ''' m2m_request:

        {u'company[]': [u'["id", 1]'],
         u'professions[]': [u'["id", 2]', u'["id", 3]']}
    '''
    m2m_request = {k: v for k, v in list(request.items()) if k.endswith('[]')}
    for k, v in list(m2m_request.items()):
        key = k[:-2]
        relation = getattr(obj.__class__, key, False)
        if not relation:
            relation = getattr(obj, key, False)
        if not relation:
            continue  # pragma: no cover
        value = get_m2m_objs(session, relation, v)
        if value is None:
            continue

        if relation.property.uselist is False:
            if value:
                value = value[0]
            else:
                value = None
        params[key] = value
    return params


class RequestPreprocessing(object):

    def __init__(self, request):
        self.request = request
        self.types_list = {'Boolean': self._check_boolean,
                           'HSTORE': self._check_hstore,
                           'JSON': self._check_hstore,
                           'JSONB': self._check_hstore,
                           'Date': self._check_date,
                           'DateTime': self._check_date,
                           'BYTEA': self._check_bytea,
                           'LargeBinary': self._check_bytea,
                           'TIMESTAMP': self._check_date
                           }

    def _check_boolean(self, value):
        if not value:
            return False
        if value in ('0', 'False', 'None', 'false'):
            return False
        return True

    def _check_bytea(self, value):
        return bytearray(value, 'utf-8')

    def _check_hstore(self, value):
        if not value:
            return None
        try:
            return json.loads(str(value))
        except Exception:
            try:
                return ast.literal_eval(str(value))
            except Exception:
                raise TypeError(
                    "[HSTORE, JSON, JSONB]: does't suppot '%s' format. %s" %
                    (value, 'Valid example: {"foo": "bar", u"baz": u"biz"}'))

    def _check_date(self, value):
        if isinstance(value, (date, datetime)):
            return value
        # XXX: I feel the dissonance here
        try:
            return datetime.strptime(value, '%Y-%m-%d')
        except ValueError:
            try:
                return datetime.strptime(value, '%Y-%m-%d %H:%M')
            except ValueError:
                return datetime.strptime(value, '%Y-%m-%d %H:%M:%S')

    def check_type(self, table, key):
        self.key = key
        self.column = None
        obj_columns = get_columns(table)
        if key in obj_columns:
            self.column = obj_columns[key]
        else:
            self.column = getattr(table, key)
        column_type = self.column.type.__class__.__name__
        value = self.request[key]
        if type(value) in (list, tuple) and\
                column_type not in ['JSON', 'JSONB', 'HSTORE']:
            value = value[0]

        if not isinstance(value, sqlalchemy.sql.elements.Null) and not value\
                and not hasattr(value, 'filename')\
                and not column_type == 'Boolean':
            if self.column.default or self.column.primary_key:
                return None

        if column_type in list(self.types_list.keys()):
            check = self.types_list[column_type]
            return check(value)

        return value


class ObjPreprocessing(object):

    def __init__(self, obj):
        self.obj = obj

    def add(self, session, data, table):
        request_preprocessing = RequestPreprocessing(data)
        # filter data for object
        for key in list(data.keys()):
            obj_columns = get_columns(self.obj)
            # check if columns not exist
            if key not in obj_columns and not hasattr(self.obj, key):
                if not key.endswith('[]'):
                    data.pop(key, None)
                continue  # pragma: no cover
            value = request_preprocessing.check_type(table, key)

            if value is None:
                data.pop(key, None)
                continue
            data[key] = value
        params = {k: v for k, v in data.items() if not k.endswith('[]')}
        m2m_params = get_m2m_value(session, data, self.obj)
        params = dict(list(params.items()) + list(m2m_params.items()))
        if inspect.isclass(self.obj):
            return self.obj(**params)
        for k, v in params.items():
            setattr(self.obj, k, v)
        return self.obj

    def delete(self):
        return self.obj



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
CREATE, READ, DELETE, UPDATE actions for SQLAlchemy models
"""
import sqlalchemy
import transaction

from .common import get_obj, get_obj_by_request_data, unjson
from .preprocessing import ObjPreprocessing


class CRUD(object):

    """ Main class for CRUD actions

        :Parameters:

            - `session`: SQLAlchemy DBSession
            - `table`: SQLAlchemy model
            - `commit`: To do or not after modify object
            - `preprocessing`: you custom preprocessing class
    """

    def __init__(self, session, table, preprocessing=ObjPreprocessing,
                 commit=True):
        self.table = table
        self.session = session
        self.commit = commit
        self.preprocessing = preprocessing

    def create(self, data, update=False, **kwargs):
        """
        Creates a new object pretreated input data.

        .. code-block:: python

            DBSession.sacrud(Users).create({'name': 'Vasya', 'sex': 1})

        Support JSON:

        .. code-block:: python

            DBSession.sacrud(Users).create('{"name": "Vasya", "sex": 1}')

        For adding multiple data for m2m or m2o use endinng `[]`, ex.:

        .. code-block:: python

            DBSession.sacrud(Users).create(
                {'name': 'Vasya', 'sex': 1,
                 'groups[]': ['["id", 1]', '["id", 2]']}
            )
        """
        data = unjson(data)

        if update is True:
            obj = get_obj_by_request_data(self.session, self.table, data)
        else:
            obj = None
        return self._add(obj, data, **kwargs)

    def read(self, *pk):
        """
        Return a list of entries in the table or single
        entry if there is an pk.

        .. code-block:: python

            # All users
            DBSession.sacrud(Users).read()

            # Composite primary_key
            DBSession.sacrud(User2Groups).read({'user_id': 4, 'group_id': 2})

            # Multiple rows
            primary_keys = [
                {'user_id': 4, 'group_id': 2},
                {'user_id': 4, 'group_id': 3},
                {'user_id': 1, 'group_id': 1},
                {'user_id': 19, 'group_id': 2}
            ]
            DBSession.sacrud(User2Groups).read(*primary_keys)

            # JSON using
            primary_keys = '''[
                {"user_id": 4, "group_id": 2},
                {"user_id": 4, "group_id": 3},
                {"user_id": 1, "group_id": 1},
                {"user_id": 19, "group_id": 2}
            ]'''
            DBSession.sacrud(User2Groups).read(primary_keys)

            # Delete
            DBSession.sacrud(User2Groups).read(*primary_keys)\
                .delete(synchronize_session=False)

            # Same, but work with only not composite primary key
            DBSession.sacrud(Users).read((5, 10))       # as list
            DBSession.sacrud(Users).read('[5, 10]')     # as JSON
            DBSession.sacrud(Users).read('{"id": 5}')   # as JSON explicit pk
            DBSession.sacrud(Users).read(5, "1", 2)     # as *args
            DBSession.sacrud(Users).read(42)            # single
        """
        pk = [unjson(obj) for obj in pk]

        if len(pk) == 1:  # like ([1,2,3,4,5], )
            return get_obj(self.session, self.table, pk[0])
        elif len(pk) > 1:  # like (1, 2, 3, 4, 5)
            return get_obj(self.session, self.table, pk)
        return self.session.query(self.table)

    def update(self, pk, data, **kwargs):
        """
        Updates the object by primary_key:

        .. code-block:: python

            DBSession.sacrud(Users).update(1, {'name': 'Petya'})
            DBSession.sacrud(Users).update('1', {'name': 'Petya'})
            DBSession.sacrud(User2Groups).update({'user_id': 4, 'group_id': 2},
                                                 {'group_id': 1})

        JSON support:

        .. code-block:: python

            DBSession.sacrud(Users).update(1, '{"name": "Petya"}')
            DBSession.sacrud(User2Groups).update(
                '{"user_id": 4, "group_id": 2}',    # primary_key
                '{"group_id": 1}'                   # data
            )

        Default it run ``session.commit() or transaction.commit()``.
        If it is not necessary use attribute ``commit=False``.
        """
        pk = unjson(pk)
        data = unjson(data)
        obj = get_obj(self.session, self.table, pk)
        return self._add(obj, data, **kwargs)

    def delete(self, pk, **kwargs):
        """
        Delete the object by primary_key:

        .. code-block:: python

            DBSession.sacrud(Users).delete(1)
            DBSession.sacrud(Users).delete('1')
            DBSession.sacrud(User2Groups).delete({'user_id': 4, 'group_id': 2})

        JSON support:

        .. code-block:: python

            DBSession.sacrud(User2Groups).delete(
                '{"user_id": 4, "group_id": 2}'
            )

        Default it run ``session.commit() or transaction.commit()``.
        If it is not necessary use attribute ``commit=False``.
        """
        pk = unjson(pk)

        obj = get_obj(self.session, self.table, pk)
        if self._delete(obj, **kwargs):
            return {'pk': pk, 'name': obj.__repr__()}

    def _add(self, obj, data, **kwargs):
        """ Update the object directly.

        .. code-block:: python

            DBSession.sacrud(Users)._add(UserObj, {'name': 'Gennady'})
        """
        if isinstance(obj, sqlalchemy.orm.query.Query):
            obj = obj.one()
        obj = self.preprocessing(obj=obj or self.table)\
            .add(self.session, data, self.table)
        self.session.add(obj)
        if kwargs.get('commit', self.commit) is True:
            try:
                self.session.commit()
            except AssertionError:
                transaction.commit()
        return obj

    def _delete(self, obj, **kwargs):
        """ Delete the object directly.

        .. code-block:: python

            DBSession.sacrud(Users)._delete(UserObj)

        If you no needed commit session

        .. code-block:: python

            DBSession.sacrud(Users, commit=False)._delete(UserObj)
        """
        if isinstance(obj, sqlalchemy.orm.query.Query):
            obj = obj.one()
        obj = self.preprocessing(obj=obj).delete()
        self.session.delete(obj)
        if kwargs.get('commit', self.commit) is True:
            try:
                self.session.commit()
            except AssertionError:
                transaction.commit()
        return True



#! /usr/bin/env python
# -*- coding: utf-8 -*-
# vim:fenc=utf-8
#
# Copyright © 2014 uralbash <root@uralbash.ru>
#
# Distributed under terms of the MIT license.

"""
SQLAlchemy helpers
"""
import json
import inspect
import itertools

import sqlalchemy
from sqlalchemy import or_, and_


def unjson(obj):
    try:
        return json.loads(obj)
    except TypeError:
        pass
    return obj


class TableProperty(object):
    def __init__(self, func):
        self.func = func

    def __get__(self, inst, cls):
        return self.func(cls.__table__)


class ClassProperty(TableProperty):
    def __get__(self, inst, cls):
        return self.func(cls)


def get_attrname_by_colname(instance, name):
    """ Get value from SQLAlchemy instance by column name

    :Parameters:
        - `instance`: SQLAlchemy model instance.
        - `name`:  Column name

    :Examples:

    >>> from sqlalchemy import Column, Integer
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> Base = declarative_base()
    >>> class MPTTPages(Base):
    ...     __tablename__ = "mptt_pages"
    ...     id = Column(Integer, primary_key=True)
    ...     left = Column("lft", Integer, nullable=False)
    >>> get_attrname_by_colname(MPTTPages(), 'lft')
    'left'
    """
    for attr, column in list(sqlalchemy.inspect(instance.__class__).c.items()):
        if column.name == name:
            return attr


def get_relationship(table):
    if not table:
        return None
    relations = sqlalchemy.inspect(table).relationships
    return [rel for rel in relations]


def get_empty_instance(table):
    """ Return  empty instance of model.
    """
    instance_defaults_params = inspect.getargspec(table.__init__).args[1:]
    # list like ['name', 'group', 'visible'] to dict with empty
    # value as {'name': None, 'group': None, 'visible': None}
    init = dict(
        list(zip(instance_defaults_params, itertools.repeat(None)))
    )
    return table(**init)


def get_obj(session, table, pk):
    if not pk:
        return None

    pk_list = get_pk(table)

    def composite_pk_clauses(pk):
        clauses = []
        for item in pk_list:
            clauses.append(getattr(table, item.name) == pk[item.name])
        return and_(*clauses)

    query = session.query(table)
    if type(pk) is list or type(pk) is tuple:
        if all(type(item) is dict for item in pk):
            clauses = []
            for item in pk:
                clauses.append(composite_pk_clauses(item))
            return query.filter(or_(*clauses))
        return query.filter(pk_list[0].in_(pk))
    elif (type(pk) is int or str(pk).isdigit()) and len(pk_list) == 1:
        return query.get(pk)
    elif type(pk) is dict:
        return query.filter(composite_pk_clauses(pk))
    return None  # pragma: no cover


def get_pk(obj):
    """ Return primary key name by model class or instance.

    :Parameters:
        - `obj`: SQLAlchemy model instance or class.

    :Examples:

    >>> from sqlalchemy import Column, Integer
    >>> from sqlalchemy.ext.declarative import declarative_base
    >>> Base = declarative_base()
    >>> class User(Base):
    ...     __tablename__ = 'users'
    ...     id = Column(Integer, primary_key=True)
    >>> get_pk(User())
    (Column('id', Integer(), table=<users>, primary_key=True, nullable=False),)
    >>> get_pk(User)
    (Column('id', Integer(), table=<users>, primary_key=True, nullable=False),)
    """
    if inspect.isclass(obj):
        pk_list = sqlalchemy.inspect(obj).primary_key
    else:
        pk_list = obj.__mapper__.primary_key
    return pk_list


def pk_to_list(obj, as_json=False):
    pk_list = []
    primary_keys = get_pk(obj)
    for item in primary_keys:
        item_name = get_attrname_by_colname(obj, item.name)
        pk_list.append(item_name)
        pk_list.append(getattr(obj, item_name))
    if as_json:
        return json.dumps(pk_list)
    return pk_list


def pk_to_dict(obj):
    pk_list = pk_to_list(obj)
    return pk_list_to_dict(pk_list)


def pk_list_to_dict(pk_list):
    """ convert list of multi pk to dict

    ['id', 1, 'id2', 22, 'foo', 'bar'] -> {'foo': 'bar', 'id2': 22, 'id': 1}
    """
    if pk_list and len(pk_list) % 2 == 0:
        return dict(zip(pk_list[::2], pk_list[1::2]))
    return None


def get_obj_by_request_data(session, table, data):
    pk = {}
    table_pk = get_pk(table)
    if not data:
        return None
    for item in table_pk:
        if item.name in data:
            pk[item.name] = data[item.name]
        else:
            return None
    try:
        return get_obj(session, table, pk)
    except sqlalchemy.orm.exc.NoResultFound:
        return None


def columns_by_group(table):
    if hasattr(table, 'sacrud_detail_col'):
        return [c for c in table.sacrud_detail_col]
    return [('', sqlalchemy.inspection.inspect(table).columns)]


def get_flat_columns(table):
    columns = []
    if not hasattr(table, 'sacrud_detail_col'):
        return columns
    for item in table.sacrud_detail_col:
        for col in item[1]:
            columns.append(col)
    return columns


def get_columns(obj):
    if hasattr(obj, '__table__'):
        return obj.__table__.columns
    return sqlalchemy.inspect(obj).mapper.columns
